@@ -124,7 +124,7 @@ void do_open_filesys(int argc, char **argv)
 	int	catastrophic = 0;
 	blk_t	superblock = 0;
 	blk_t	blocksize = 0;
-	int	open_flags = 0;
+	int	open_flags = EXT2_FLAG_SOFTSUPP_FEATURES;
 	char	*data_filename = 0;
 	
 	reset_getopt();
@@ -181,9 +181,10 @@ print_usage:
 
 void do_lcd(int argc, char **argv)
 {
-	if (common_args_process(argc, argv, 2, 2, "lcd",
-				"<native dir>", 0))
+	if (argc != 2) {
+		com_err(argv[0], 0, "Usage: %s %s", argv[0], "<native dir>");
 		return;
+	}
 
 	if (chdir(argv[1]) == -1) {
 		com_err(argv[0], errno,
@@ -377,10 +378,12 @@ static void finish_range(struct list_blocks_struct *lb)
 	else
 		fprintf(lb->f, ", ");
 	if (lb->first_block == lb->last_block)
-		fprintf(lb->f, "(%lld):%u", lb->first_bcnt, lb->first_block);
+		fprintf(lb->f, "(%lld):%u",
+			(long long)lb->first_bcnt, lb->first_block);
 	else
-		fprintf(lb->f, "(%lld-%lld):%u-%u", lb->first_bcnt,
-			lb->last_bcnt, lb->first_block, lb->last_block);
+		fprintf(lb->f, "(%lld-%lld):%u-%u",
+			(long long)lb->first_bcnt, (long long)lb->last_bcnt,
+			lb->first_block, lb->last_block);
 	lb->first_block = 0;
 }
 
@@ -431,25 +434,28 @@ static int list_blocks_proc(ext2_filsys fs EXT2FS_ATTR((unused)),
 
 static void dump_xattr_string(FILE *out, const char *str, int len)
 {
-	int printable = 1;
+	int printable = 0;
 	int i;
 	
-	/* check is string printable? */
+	/* check: is string "printable enough?" */
 	for (i = 0; i < len; i++)
-		if (!isprint(str[i])) {
-			printable = 0;
-			break;
-		}
+		if (isprint(str[i]))
+			printable++;
+
+	if (printable <= len*7/8)
+		printable = 0;
 
 	for (i = 0; i < len; i++)
 		if (printable)
-			fprintf(out, "%c", str[i]);
+			fprintf(out, isprint(str[i]) ? "%c" : "\\%03o",
+				(unsigned char)str[i]);
 		else
-			fprintf(out, "%02x ", str[i]);
+			fprintf(out, "%02x ", (unsigned char)str[i]);
 }
 
-static void internal_dump_inode_extra(FILE *out, const char *prefix,
-				      ext2_ino_t inode_num, 
+static void internal_dump_inode_extra(FILE *out, 
+				      const char *prefix EXT2FS_ATTR((unused)),
+				      ext2_ino_t inode_num EXT2FS_ATTR((unused)), 
 				      struct ext2_inode_large *inode)
 {
 	struct ext2_ext_attr_entry *entry;
@@ -507,7 +513,7 @@ static void dump_blocks(FILE *f, const char *prefix, ext2_ino_t inode)
 			     list_blocks_proc, (void *)&lb);
 	finish_range(&lb);
 	if (lb.total)
-		fprintf(f, "\n%sTOTAL: %lld\n", prefix, lb.total);
+		fprintf(f, "\n%sTOTAL: %lld\n", prefix, (long long)lb.total);
 	fprintf(f,"\n");
 }
 
@@ -533,15 +539,15 @@ void internal_dump_inode(FILE *out, const char *prefix,
 		prefix, 
 		inode->i_mode & 0777, inode->i_flags, inode->i_generation);
 	fprintf(out, "%sUser: %5d   Group: %5d   Size: ",
-		prefix, inode->i_uid, inode->i_gid);
+		prefix, inode_uid(*inode), inode_gid(*inode));
 	if (LINUX_S_ISREG(inode->i_mode)) {
-		__u64 i_size = (inode->i_size |
-				((unsigned long long)inode->i_size_high << 32));
+		unsigned long long i_size = (inode->i_size |
+				    ((unsigned long long)inode->i_size_high << 32));
 
-		fprintf(out, "%lld\n", i_size);
+		fprintf(out, "%llu\n", i_size);
 	} else
 		fprintf(out, "%d\n", inode->i_size);
-	if (current_fs->super->s_creator_os == EXT2_OS_HURD)
+	if (os == EXT2_OS_HURD)
 		fprintf(out,
 			"%sFile ACL: %d    Directory ACL: %d Translator: %d\n",
 			prefix,
@@ -551,13 +557,16 @@ void internal_dump_inode(FILE *out, const char *prefix,
 		fprintf(out, "%sFile ACL: %d    Directory ACL: %d\n",
 			prefix,
 			inode->i_file_acl, LINUX_S_ISDIR(inode->i_mode) ? inode->i_dir_acl : 0);
-	fprintf(out, "%sLinks: %d   Blockcount: %u\n",
-		prefix, inode->i_links_count, inode->i_blocks);
+	if (os == EXT2_OS_LINUX) 
+		fprintf(out, "%sLinks: %d   Blockcount: %llu\n",
+			prefix, inode->i_links_count, 
+			(((unsigned long long) 
+			  inode->osd2.linux2.l_i_blocks_hi << 32)) + 
+			inode->i_blocks);
+	else
+		fprintf(out, "%sLinks: %d   Blockcount: %u\n",
+			prefix, inode->i_links_count, inode->i_blocks);
 	switch (os) {
-	    case EXT2_OS_LINUX:
-		frag = inode->osd2.linux2.l_i_frag;
-		fsize = inode->osd2.linux2.l_i_fsize;
-		break;
 	    case EXT2_OS_HURD:
 		frag = inode->osd2.hurd2.h_i_frag;
 		fsize = inode->osd2.hurd2.h_i_fsize;
@@ -722,7 +731,7 @@ void do_testi(int argc, char *argv[])
 void do_freeb(int argc, char *argv[])
 {
 	blk_t block;
-	int count = 1;
+	blk_t count = 1;
 
 	if (common_block_args_process(argc, argv, &block, &count))
 		return;
@@ -741,7 +750,7 @@ void do_freeb(int argc, char *argv[])
 void do_setb(int argc, char *argv[])
 {
 	blk_t block;
-	int count = 1;
+	blk_t count = 1;
 
 	if (common_block_args_process(argc, argv, &block, &count))
 		return;
@@ -760,7 +769,7 @@ void do_setb(int argc, char *argv[])
 void do_testb(int argc, char *argv[])
 {
 	blk_t block;
-	int count = 1;
+	blk_t count = 1;
 
 	if (common_block_args_process(argc, argv, &block, &count))
 		return;
@@ -867,6 +876,9 @@ void do_modify_inode(int argc, char *argv[])
 	modify_u32(argv[0], "Access time", decimal_format, &inode.i_atime);
 	modify_u32(argv[0], "Deletion time", decimal_format, &inode.i_dtime);
 	modify_u16(argv[0], "Link count", decimal_format, &inode.i_links_count);
+	if (os == EXT2_OS_LINUX)
+		modify_u16(argv[0], "Block count high", unsignedlong_format, 
+			   &inode.osd2.linux2.l_i_blocks_hi);
 	modify_u32(argv[0], "Block count", unsignedlong_format, &inode.i_blocks);
 	modify_u32(argv[0], "File flags", hex_format, &inode.i_flags);
 	modify_u32(argv[0], "Generation", hex_format, &inode.i_generation);
@@ -879,16 +891,12 @@ void do_modify_inode(int argc, char *argv[])
 	else
 		modify_u32(argv[0], "High 32bits of size", decimal_format, &inode.i_size_high);
 
-	if (current_fs->super->s_creator_os == EXT2_OS_HURD)
+	if (os == EXT2_OS_HURD)
 		modify_u32(argv[0], "Translator Block",
 			    decimal_format, &inode.osd1.hurd1.h_i_translator);
 	
 	modify_u32(argv[0], "Fragment address", decimal_format, &inode.i_faddr);
 	switch (os) {
-	    case EXT2_OS_LINUX:
-		frag = &inode.osd2.linux2.l_i_frag;
-		fsize = &inode.osd2.linux2.l_i_fsize;
-		break;
 	    case EXT2_OS_HURD:
 		frag = &inode.osd2.hurd2.h_i_frag;
 		fsize = &inode.osd2.hurd2.h_i_fsize;
@@ -950,15 +958,23 @@ void do_print_working_directory(int argc, char *argv[])
 		com_err(argv[0], retval,
 			"while trying to get pathname of cwd");
 	}
-	printf("[pwd]   INODE: %6u  PATH: %s\n", cwd, pathname);
-	free(pathname);
+	printf("[pwd]   INODE: %6u  PATH: %s\n",
+	       cwd, pathname ? pathname : "NULL");
+        if (pathname) {
+		free(pathname);
+		pathname = NULL;
+        }
 	retval = ext2fs_get_pathname(current_fs, root, 0, &pathname);
 	if (retval) {
 		com_err(argv[0], retval,
 			"while trying to get pathname of root");
 	}
-	printf("[root]  INODE: %6u  PATH: %s\n", root, pathname);
-	free(pathname);
+	printf("[root]  INODE: %6u  PATH: %s\n",
+	       root, pathname ? pathname : "NULL");
+	if (pathname) {
+		free(pathname);
+		pathname = NULL;
+	}
 	return;
 }
 
@@ -1070,8 +1086,8 @@ void do_undel(int argc, char *argv[])
 	ext2_ino_t	ino;
 	struct ext2_inode inode;
 
-	if (common_args_process(argc, argv, 3, 3, "undelete",
-				"<inode_num> <dest_name>",
+	if (common_args_process(argc, argv, 2, 3, "undelete",
+				"<inode_num> [dest_name]",
 				CHECK_FS_RW | CHECK_FS_BITMAPS))
 		return;
 
@@ -1102,7 +1118,8 @@ void do_undel(int argc, char *argv[])
 
 	ext2fs_inode_alloc_stats2(current_fs, ino, +1, 0);
 
-	make_link(argv[1], argv[2]);
+	if (argc > 2)
+		make_link(argv[1], argv[2]);
 }
 
 static void unlink_file_by_name(char *filename)
@@ -1138,7 +1155,7 @@ void do_unlink(int argc, char *argv[])
 
 void do_find_free_block(int argc, char *argv[])
 {
-	blk_t	free_blk, goal;
+	blk_t	free_blk, goal, first_free = 0;
  	int		count;
 	errcode_t	retval;
 	char		*tmp;
@@ -1174,6 +1191,11 @@ void do_find_free_block(int argc, char *argv[])
 	while (count-- > 0) {
 		retval = ext2fs_new_block(current_fs, free_blk + 1, 0,
 					  &free_blk);
+		if (first_free) {
+			if (first_free == free_blk)
+				break;
+		} else
+			first_free = free_blk;
 		if (retval) {
 			com_err("ext2fs_new_block", retval, 0);
 			return;
@@ -1790,7 +1812,7 @@ int main(int argc, char **argv)
 	int		sci_idx;
 	const char	*usage = "Usage: debugfs [-b blocksize] [-s superblock] [-f cmd_file] [-R request] [-V] [[-w] [-c] device]";
 	int		c;
-	int		open_flags = 0;
+	int		open_flags = EXT2_FLAG_SOFTSUPP_FEATURES;
 	char		*request = 0;
 	int		exit_status = 0;
 	char		*cmd_file = 0;
@@ -1799,7 +1821,7 @@ int main(int argc, char **argv)
 	int		catastrophic = 0;
 	char		*data_filename = 0;
 	
-	initialize_ext2_error_table();
+	add_error_table(&et_ext2_error_table);
 	fprintf (stderr, "debugfs %s (%s)\n", E2FSPROGS_VERSION,
 		 E2FSPROGS_DATE);
 
@@ -1875,5 +1897,6 @@ int main(int argc, char **argv)
 	if (current_fs)
 		close_filesystem();
 	
+	remove_error_table(&et_ext2_error_table);
 	return exit_status;
 }
