@@ -175,6 +175,27 @@ struct problem_context {
 /* Backup journal inode blocks */
 #define PR_0_BACKUP_JNL				0x00002D
 
+/* Reserved blocks w/o resize_inode */
+#define PR_0_NONZERO_RESERVED_GDT_BLOCKS	0x00002E
+
+/* Resize_inode not enabled, but resize inode is non-zero */
+#define PR_0_CLEAR_RESIZE_INODE			0x00002F
+
+/* Resize inode invalid */
+#define PR_0_RESIZE_INODE_INVALID		0x000030
+
+/* Last mount time is in the future */
+#define PR_0_FUTURE_SB_LAST_MOUNT		0x000031
+
+/* Last write time is in the future */
+#define PR_0_FUTURE_SB_LAST_WRITE		0x000032
+
+/* Superblock hint for external journal incorrect */
+#define PR_0_EXTERNAL_JOURNAL_HINT		0x000033
+
+/* Superblock hint for external journal incorrect */
+#define PR_0_DIRHASH_HINT			0x000034
+
 /*
  * Pass 1 errors
  */
@@ -410,6 +431,30 @@ struct problem_context {
 /* Bad block has indirect block that conflicts with filesystem block */
 #define PR_1_BB_FS_BLOCK		0x01004D
 
+/* Resize inode failed */
+#define PR_1_RESIZE_INODE_CREATE	0x01004E
+
+/* inode->i_size is too long */	
+#define PR_1_EXTRA_ISIZE		0x01004F
+
+/* attribute name is too long */
+#define PR_1_ATTR_NAME_LEN		0x010050
+
+/* wrong EA value offset */
+#define PR_1_ATTR_VALUE_OFFSET		0x010051
+
+/* wrong EA blocknumber */
+#define PR_1_ATTR_VALUE_BLOCK		0x010052
+
+/* wrong EA value size */
+#define PR_1_ATTR_VALUE_SIZE		0x010053
+
+/* wrong EA hash value */
+#define PR_1_ATTR_HASH			0x010054
+
+/* inode appears to be a directory */
+#define PR_1_TREAT_AS_DIRECTORY		0x010055
+
 /*
  * Pass 1b errors
  */
@@ -628,7 +673,7 @@ struct problem_context {
 #define PR_2_HTREE_CLEAR	0x020038
 
 /* Clear the htree flag forcibly */
-#define PR_2_HTREE_FCLR		0x020039
+/* #define PR_2_HTREE_FCLR	0x020039 */
 
 /* Bad block in htree interior node */
 #define PR_2_HTREE_BADBLK	0x02003A
@@ -660,6 +705,12 @@ struct problem_context {
 /* Duplicate directory entry found */
 #define PR_2_REPORT_DUP_DIRENT	0x020043
 
+/* i_blocks_hi should be zero */
+#define PR_2_BLOCKS_HI_ZERO	0x020044
+
+/* Unexpected HTREE block */
+#define PR_2_UNEXPECTED_HTREE_BLOCK	0x020045
+
 /*
  * Pass 3 errors
  */
@@ -852,6 +903,13 @@ struct problem_context {
 /* Inode rangeused, but not marked used in bitmap */
 #define PR_5_INODE_RANGE_USED		0x050017
 
+/*
+ * Post-Pass 5 errors
+ */
+
+/* Recreate the journal if E2F_FLAG_JOURNAL_INODE flag is set */
+#define PR_6_RECREATE_JOURNAL		0x060001
+
 /*
  * Function declarations
  */
@@ -863,5 +921,6 @@ void clear_problem_context(struct problem_context *ctx);
 
 /* message.c */
 void print_e2fsck_message(e2fsck_t ctx, const char *msg,
-			  struct problem_context *pctx, int first);
+			  struct problem_context *pctx, int first, 
+			  int recurse);
 
