@@ -108,7 +108,7 @@ errcode_t resize_fs(ext2_filsys fs, blk_t *new_size, int flags,
 
 #ifdef RESIZE2FS_DEBUG
 	if (rfs->flags & RESIZE_DEBUG_BMOVE)
-		printf("Number of free blocks: %d/%d, Needed: %d\n",
+		printf("Number of free blocks: %u/%u, Needed: %d\n",
 		       rfs->old_fs->super->s_free_blocks_count,
 		       rfs->new_fs->super->s_free_blocks_count,
 		       rfs->needed_blocks);
@@ -138,6 +138,7 @@ errcode_t resize_fs(ext2_filsys fs, blk_t *new_size, int flags,
 	if (retval)
 		goto errout;
 
+	rfs->new_fs->flags &= ~EXT2_FLAG_MASTER_SB_ONLY;	
 	retval = ext2fs_close(rfs->new_fs);
 	if (retval)
 		goto errout;
@@ -172,38 +173,32 @@ errout:
  */
 
 /*
- * This routine adjusts the superblock and other data structures...
+ * This routine is shared by the online and offline resize routines.
+ * All of the information which is adjusted in memory is done here.
  */
-static errcode_t adjust_superblock(ext2_resize_t rfs, blk_t new_size)
+errcode_t adjust_fs_info(ext2_filsys fs, ext2_filsys old_fs, blk_t new_size)
 {
-	ext2_filsys fs;
-	int		overhead = 0;
-	int		rem, adj = 0;
 	errcode_t	retval;
-	ext2_ino_t	real_end;
+	int		overhead = 0;
+	int		rem;
 	blk_t		blk, group_block;
-	unsigned long	i, j, old_desc_blocks;
-	int		old_numblocks, numblocks, adjblocks;
+	ext2_ino_t	real_end;
+	int		adj, old_numblocks, numblocks, adjblocks;
+	unsigned long	i, j, old_desc_blocks, max_group;
 	unsigned int	meta_bg, meta_bg_size;
 	int		has_super;
-	unsigned long	max_group;
-	
-	fs = rfs->new_fs;
+	unsigned long long new_inodes;	/* u64 to check for overflow */
+
 	fs->super->s_blocks_count = new_size;
-	ext2fs_mark_super_dirty(fs);
-	ext2fs_mark_bb_dirty(fs);
-	ext2fs_mark_ib_dirty(fs);
 
 retry:
-	fs->group_desc_count = (fs->super->s_blocks_count -
-				fs->super->s_first_data_block +
-				EXT2_BLOCKS_PER_GROUP(fs->super) - 1)
-		/ EXT2_BLOCKS_PER_GROUP(fs->super);
+	fs->group_desc_count = ext2fs_div_ceil(fs->super->s_blocks_count -
+				       fs->super->s_first_data_block,
+				       EXT2_BLOCKS_PER_GROUP(fs->super));
 	if (fs->group_desc_count == 0)
 		return EXT2_ET_TOOSMALL;
-	fs->desc_blocks = (fs->group_desc_count +
-			   EXT2_DESC_PER_BLOCK(fs->super) - 1)
-		/ EXT2_DESC_PER_BLOCK(fs->super);
+	fs->desc_blocks = ext2fs_div_ceil(fs->group_desc_count, 
+					  EXT2_DESC_PER_BLOCK(fs->super));
 
 	/*
 	 * Overhead is the number of bookkeeping blocks per group.  It
@@ -233,13 +228,19 @@ retry:
 	/*
 	 * Adjust the number of inodes
 	 */
+	new_inodes =(unsigned long long) fs->super->s_inodes_per_group * fs->group_desc_count;
+	if (new_inodes > ~0U) {
+		fprintf(stderr, _("inodes (%llu) must be less than %u"),
+				   new_inodes, ~0U);
+		return EXT2_ET_TOO_MANY_INODES;
+	}
 	fs->super->s_inodes_count = fs->super->s_inodes_per_group *
 		fs->group_desc_count;
 
 	/*
 	 * Adjust the number of free blocks
 	 */
-	blk = rfs->old_fs->super->s_blocks_count;
+	blk = old_fs->super->s_blocks_count;
 	if (blk > fs->super->s_blocks_count)
 		fs->super->s_free_blocks_count -=
 			(blk - fs->super->s_blocks_count);
@@ -250,10 +251,10 @@ retry:
 	/*
 	 * Adjust the number of reserved blocks
 	 */
-	blk = rfs->old_fs->super->s_r_blocks_count * 100 /
-		rfs->old_fs->super->s_blocks_count;
-	fs->super->s_r_blocks_count = ((fs->super->s_blocks_count * blk)
-				       / 100);
+	blk = (__u64)old_fs->super->s_r_blocks_count * 100 /
+		old_fs->super->s_blocks_count;
+	fs->super->s_r_blocks_count = e2p_percent(blk, 
+						  fs->super->s_blocks_count);
 
 	/*
 	 * Adjust the bitmaps for size
@@ -274,23 +275,18 @@ retry:
 	/*
 	 * Reallocate the group descriptors as necessary.
 	 */
-	if (rfs->old_fs->desc_blocks != fs->desc_blocks) {
-		retval = ext2fs_resize_mem(rfs->old_fs->desc_blocks *
+	if (old_fs->desc_blocks != fs->desc_blocks) {
+		retval = ext2fs_resize_mem(old_fs->desc_blocks *
 					   fs->blocksize,
 					   fs->desc_blocks * fs->blocksize,
 					   &fs->group_desc);
 		if (retval)
 			goto errout;
-	}
-
-	/*
-	 * Check to make sure there are enough inodes
-	 */
-	if ((rfs->old_fs->super->s_inodes_count -
-	     rfs->old_fs->super->s_free_inodes_count) >
-	    rfs->new_fs->super->s_inodes_count) {
-		retval = ENOSPC;
-		goto errout;
+		if (fs->desc_blocks > old_fs->desc_blocks) 
+			memset((char *) fs->group_desc + 
+			       (old_fs->desc_blocks * fs->blocksize), 0,
+			       (fs->desc_blocks - old_fs->desc_blocks) *
+			       fs->blocksize);
 	}
 
 	/*
@@ -301,14 +297,14 @@ retry:
 	 */
 	if ((fs->super->s_feature_compat & 
 	     EXT2_FEATURE_COMPAT_RESIZE_INODE) &&
-	    (rfs->old_fs->desc_blocks != fs->desc_blocks)) {
+	    (old_fs->desc_blocks != fs->desc_blocks)) {
 		int new;
 
 		new = ((int) fs->super->s_reserved_gdt_blocks) + 
-			(rfs->old_fs->desc_blocks - fs->desc_blocks);
+			(old_fs->desc_blocks - fs->desc_blocks);
 		if (new < 0)
 			new = 0;
-		if (new > fs->blocksize/4)
+		if (new > (int) fs->blocksize/4)
 			new = fs->blocksize/4;
 		fs->super->s_reserved_gdt_blocks = new;
 		if (new == 0)
@@ -320,27 +316,28 @@ retry:
 	 * If we are shrinking the number block groups, we're done and
 	 * can exit now.
 	 */
-	if (rfs->old_fs->group_desc_count > fs->group_desc_count) {
+	if (old_fs->group_desc_count > fs->group_desc_count) {
 		retval = 0;
 		goto errout;
 	}
+
 	/*
 	 * Fix the count of the last (old) block group
 	 */
-	old_numblocks = (rfs->old_fs->super->s_blocks_count -
-			 rfs->old_fs->super->s_first_data_block) %
-				 rfs->old_fs->super->s_blocks_per_group;
+	old_numblocks = (old_fs->super->s_blocks_count -
+			 old_fs->super->s_first_data_block) %
+				 old_fs->super->s_blocks_per_group;
 	if (!old_numblocks)
-		old_numblocks = rfs->old_fs->super->s_blocks_per_group;
-	if (rfs->old_fs->group_desc_count == fs->group_desc_count) {
-		numblocks = (rfs->new_fs->super->s_blocks_count -
-			     rfs->new_fs->super->s_first_data_block) %
-				     rfs->new_fs->super->s_blocks_per_group;
+		old_numblocks = old_fs->super->s_blocks_per_group;
+	if (old_fs->group_desc_count == fs->group_desc_count) {
+		numblocks = (fs->super->s_blocks_count -
+			     fs->super->s_first_data_block) %
+			fs->super->s_blocks_per_group;
 		if (!numblocks)
-			numblocks = rfs->new_fs->super->s_blocks_per_group;
+			numblocks = fs->super->s_blocks_per_group;
 	} else
-		numblocks = rfs->new_fs->super->s_blocks_per_group;
-	i = rfs->old_fs->group_desc_count - 1;
+		numblocks = fs->super->s_blocks_per_group;
+	i = old_fs->group_desc_count - 1;
 	fs->group_desc[i].bg_free_blocks_count += (numblocks-old_numblocks);
 		
 	/*
@@ -348,36 +345,25 @@ retry:
 	 * done and can exit now.  (If the number block groups is
 	 * shrinking, we had exited earlier.)
 	 */
-	if (rfs->old_fs->group_desc_count >= fs->group_desc_count) {
+	if (old_fs->group_desc_count >= fs->group_desc_count) {
 		retval = 0;
 		goto errout;
 	}
+
 	/*
 	 * Initialize the new block group descriptors
 	 */
-	retval = ext2fs_get_mem(fs->blocksize * fs->inode_blocks_per_group,
-				&rfs->itable_buf);
-	if (retval)
-		goto errout;
-
-	memset(rfs->itable_buf, 0, fs->blocksize * fs->inode_blocks_per_group);
 	group_block = fs->super->s_first_data_block +
-		rfs->old_fs->group_desc_count * fs->super->s_blocks_per_group;
+		old_fs->group_desc_count * fs->super->s_blocks_per_group;
 
-	adj = rfs->old_fs->group_desc_count;
+	adj = old_fs->group_desc_count;
 	max_group = fs->group_desc_count - adj;
-	if (rfs->progress) {
-		retval = rfs->progress(rfs, E2_RSZ_EXTEND_ITABLE_PASS,
-				       0, max_group);
-		if (retval)
-			goto errout;
-	}
 	if (fs->super->s_feature_incompat & EXT2_FEATURE_INCOMPAT_META_BG)
 		old_desc_blocks = fs->super->s_first_meta_bg;
 	else
 		old_desc_blocks = fs->desc_blocks + 
 			fs->super->s_reserved_gdt_blocks;
-	for (i = rfs->old_fs->group_desc_count;
+	for (i = old_fs->group_desc_count;
 	     i < fs->group_desc_count; i++) {
 		memset(&fs->group_desc[i], 0,
 		       sizeof(struct ext2_group_desc));
@@ -433,6 +419,87 @@ retry:
 		retval = ext2fs_allocate_group_table(fs, i, 0);
 		if (retval) goto errout;
 
+		group_block += fs->super->s_blocks_per_group;
+	}
+	retval = 0;
+
+errout:
+	return (retval);
+}
+
+/*
+ * This routine adjusts the superblock and other data structures, both
+ * in disk as well as in memory...
+ */
+static errcode_t adjust_superblock(ext2_resize_t rfs, blk_t new_size)
+{
+	ext2_filsys fs;
+	int		adj = 0;
+	errcode_t	retval;
+	blk_t		group_block;
+	unsigned long	i;
+	unsigned long	max_group;
+	
+	fs = rfs->new_fs;
+	ext2fs_mark_super_dirty(fs);
+	ext2fs_mark_bb_dirty(fs);
+	ext2fs_mark_ib_dirty(fs);
+
+	retval = adjust_fs_info(fs, rfs->old_fs, new_size);
+	if (retval)
+		goto errout;
+
+	/*
+	 * Check to make sure there are enough inodes
+	 */
+	if ((rfs->old_fs->super->s_inodes_count -
+	     rfs->old_fs->super->s_free_inodes_count) >
+	    rfs->new_fs->super->s_inodes_count) {
+		retval = ENOSPC;
+		goto errout;
+	}
+
+	/*
+	 * If we are shrinking the number block groups, we're done and
+	 * can exit now.
+	 */
+	if (rfs->old_fs->group_desc_count > fs->group_desc_count) {
+		retval = 0;
+		goto errout;
+	}
+
+	/*
+	 * If the number of block groups is staying the same, we're
+	 * done and can exit now.  (If the number block groups is
+	 * shrinking, we had exited earlier.)
+	 */
+	if (rfs->old_fs->group_desc_count >= fs->group_desc_count) {
+		retval = 0;
+		goto errout;
+	}
+
+	/*
+	 * Initialize the new block group descriptors
+	 */
+	retval = ext2fs_get_array(fs->blocksize, fs->inode_blocks_per_group,
+				&rfs->itable_buf);
+	if (retval)
+		goto errout;
+
+	memset(rfs->itable_buf, 0, fs->blocksize * fs->inode_blocks_per_group);
+	group_block = fs->super->s_first_data_block +
+		rfs->old_fs->group_desc_count * fs->super->s_blocks_per_group;
+
+	adj = rfs->old_fs->group_desc_count;
+	max_group = fs->group_desc_count - adj;
+	if (rfs->progress) {
+		retval = rfs->progress(rfs, E2_RSZ_EXTEND_ITABLE_PASS,
+				       0, max_group);
+		if (retval)
+			goto errout;
+	}
+	for (i = rfs->old_fs->group_desc_count;
+	     i < fs->group_desc_count; i++) {
 		/*
 		 * Write out the new inode table
 		 */
@@ -477,16 +544,13 @@ errout:
 static errcode_t mark_table_blocks(ext2_filsys fs,
 				   ext2fs_block_bitmap bmap)
 {
-	blk_t			block, b;
+	blk_t			b;
 	unsigned int		j;
 	dgrp_t			i;
-	unsigned long		meta_bg, meta_bg_size;
-	int			has_super;
+	unsigned long		meta_bg_size;
 	unsigned int		old_desc_blocks;
-	errcode_t		retval;
 
 	meta_bg_size = (fs->blocksize / sizeof (struct ext2_group_desc));
-	block = fs->super->s_first_data_block;
 	if (fs->super->s_feature_incompat & EXT2_FEATURE_INCOMPAT_META_BG)
 		old_desc_blocks = fs->super->s_first_meta_bg;
 	else
@@ -514,7 +578,6 @@ static errcode_t mark_table_blocks(ext2_filsys fs,
 		 */
 		ext2fs_mark_block_bitmap(bmap,
 					 fs->group_desc[i].bg_inode_bitmap);
-		block += fs->super->s_blocks_per_group;
 	}
 	return 0;
 }
@@ -854,7 +917,7 @@ static errcode_t block_mover(ext2_resize_t rfs)
 
 	new_blk = fs->super->s_first_data_block;
 	if (!rfs->itable_buf) {
-		retval = ext2fs_get_mem(fs->blocksize *
+		retval = ext2fs_get_array(fs->blocksize,
 					fs->inode_blocks_per_group,
 					&rfs->itable_buf);
 		if (retval)
@@ -1046,13 +1109,14 @@ static errcode_t inode_scan_and_fix(ext2_resize_t rfs)
 {
 	struct process_block_struct	pb;
 	ext2_ino_t		ino, new_inode;
-	struct ext2_inode 	inode;
+	struct ext2_inode 	*inode = NULL;
 	ext2_inode_scan 	scan = NULL;
 	errcode_t		retval;
 	int			group;
 	char			*block_buf = 0;
 	ext2_ino_t		start_to_move;
 	blk_t			orig_size, new_block;
+	int			inode_size;
 	
 	if ((rfs->old_fs->group_desc_count <=
 	     rfs->new_fs->group_desc_count) &&
@@ -1075,7 +1139,7 @@ static errcode_t inode_scan_and_fix(ext2_resize_t rfs)
 
 	retval = ext2fs_init_dblist(rfs->old_fs, 0);
 	if (retval) goto errout;
-	retval = ext2fs_get_mem(rfs->old_fs->blocksize * 3, &block_buf);
+	retval = ext2fs_get_array(rfs->old_fs->blocksize, 3, &block_buf);
 	if (retval) goto errout;
 
 	start_to_move = (rfs->new_fs->group_desc_count *
@@ -1089,37 +1153,43 @@ static errcode_t inode_scan_and_fix(ext2_resize_t rfs)
 	}
 	ext2fs_set_inode_callback(scan, progress_callback, (void *) rfs);
 	pb.rfs = rfs;
-	pb.inode = &inode;
+	pb.inode = inode;
 	pb.error = 0;
 	new_inode = EXT2_FIRST_INODE(rfs->new_fs->super);
+	inode_size = EXT2_INODE_SIZE(rfs->new_fs->super);
+	inode = malloc(inode_size);
+	if (!inode) {
+		retval = ENOMEM;
+		goto errout;
+	}
 	/*
 	 * First, copy all of the inodes that need to be moved
 	 * elsewhere in the inode table
 	 */
 	while (1) {
-		retval = ext2fs_get_next_inode(scan, &ino, &inode);
+		retval = ext2fs_get_next_inode_full(scan, &ino, inode, inode_size);
 		if (retval) goto errout;
 		if (!ino)
 			break;
 
-		if (inode.i_links_count == 0 && ino != EXT2_RESIZE_INO)
+		if (inode->i_links_count == 0 && ino != EXT2_RESIZE_INO)
 			continue; /* inode not in use */
 
-		pb.is_dir = LINUX_S_ISDIR(inode.i_mode);
+		pb.is_dir = LINUX_S_ISDIR(inode->i_mode);
 		pb.changed = 0;
 
-		if (inode.i_file_acl && rfs->bmap) {
+		if (inode->i_file_acl && rfs->bmap) {
 			new_block = ext2fs_extent_translate(rfs->bmap, 
-							    inode.i_file_acl);
+							    inode->i_file_acl);
 			if (new_block) {
-				inode.i_file_acl = new_block;
-				retval = ext2fs_write_inode(rfs->old_fs, 
-							    ino, &inode);
+				inode->i_file_acl = new_block;
+				retval = ext2fs_write_inode_full(rfs->old_fs, 
+							    ino, inode, inode_size);
 				if (retval) goto errout;
 			}
 		}
 		
-		if (ext2fs_inode_has_valid_blocks(&inode) &&
+		if (ext2fs_inode_has_valid_blocks(inode) &&
 		    (rfs->bmap || pb.is_dir)) {
 			pb.ino = ino;
 			retval = ext2fs_block_iterate2(rfs->old_fs,
@@ -1152,15 +1222,17 @@ static errcode_t inode_scan_and_fix(ext2_resize_t rfs)
 		ext2fs_mark_inode_bitmap(rfs->new_fs->inode_map, new_inode);
 		if (pb.changed) {
 			/* Get the new version of the inode */
-			retval = ext2fs_read_inode(rfs->old_fs, ino, &inode);
+			retval = ext2fs_read_inode_full(rfs->old_fs, ino,
+						inode, inode_size);
 			if (retval) goto errout;
 		}
-		inode.i_ctime = time(0);
-		retval = ext2fs_write_inode(rfs->old_fs, new_inode, &inode);
+		inode->i_ctime = time(0);
+		retval = ext2fs_write_inode_full(rfs->old_fs, new_inode,
+						inode, inode_size);
 		if (retval) goto errout;
 
 		group = (new_inode-1) / EXT2_INODES_PER_GROUP(rfs->new_fs->super);
-		if (LINUX_S_ISDIR(inode.i_mode))
+		if (LINUX_S_ISDIR(inode->i_mode))
 			rfs->new_fs->group_desc[group].bg_used_dirs_count++;
 		
 #ifdef RESIZE2FS_DEBUG
@@ -1186,6 +1258,8 @@ errout:
 		ext2fs_close_inode_scan(scan);
 	if (block_buf)
 		ext2fs_free_mem(&block_buf);
+	if (inode)
+		free(inode);
 	return retval;
 }
 
@@ -1244,7 +1318,9 @@ static int check_and_change_inodes(ext2_ino_t dir,
 	retval = ext2fs_read_inode(is->rfs->old_fs, dir, &inode);
 	if (retval == 0) {
 		inode.i_mtime = inode.i_ctime = time(0);
-		ext2fs_write_inode(is->rfs->old_fs, dir, &inode);
+		is->err = ext2fs_write_inode(is->rfs->old_fs, dir, &inode);
+		if (is->err)
+			return DIRENT_ABORT;
 	}
 
 	return DIRENT_CHANGED;
@@ -1379,7 +1455,7 @@ static errcode_t move_itables(ext2_resize_t rfs)
 		 * things by not rewriting blocks that we know to be zero
 		 * already.
 		 */
-		for (cp = rfs->itable_buf+size, n=0; n < size; n++, cp--)
+		for (cp = rfs->itable_buf+size-1, n=0; n < size; n++, cp--)
 			if (*cp)
 				break;
 		n = n >> EXT2_BLOCK_SIZE_BITS(fs->super);
@@ -1525,7 +1601,9 @@ static errcode_t ext2fs_calculate_summary_stats(ext2_filsys fs)
 	total_free = 0;
 	count = 0;
 	group = 0;
-	for (ino = 1; ino <= fs->super->s_inodes_count; ino++) {
+
+	/* Protect loop from wrap-around if s_inodes_count maxed */
+	for (ino = 1; ino <= fs->super->s_inodes_count && ino > 0; ino++) {
 		if (!ext2fs_fast_test_inode_bitmap(fs->inode_map, ino)) {
 			group_free++;
 			total_free++;
